/************************************************************************
	superlu_dbg_v1.c  -- Column-Compressed-Format
	Debugging superlu -- debug version 1.
	10-31-02

	Copied from ../EXAMPLE/superlu.c
	Modified by G. Shi, 10-23-02, for testing some circuit systems.
	Data (A,b) are provided by files in the format of (row, col, value).
*************************************************************************/

/*
 * -- SuperLU routine (version 2.0) --
 * Univ. of California Berkeley, Xerox Palo Alto Research Center,
 * and Lawrence Berkeley National Lab.
 * November 15, 1997
 *
 */
#include "dsp_defs.h"
#include "util.h"
#include "time.h"
#include <stdlib.h>
#include <assert.h>

#define		HZ		CLK_TCK

#define		DEBUG	1

void   xx_preorder(char *refact,  SuperMatrix *A, int *perm_c, 
	    	int *etree, SuperMatrix *AC);

int    *xxTreePostorder( int n, int *parent);


double	Time()
{
	struct time {long  user, system, childuser, childsystem;} time;

	(void) times(&time);
	return  (double) time.user / (double) HZ;
}


/*** Globals ***/
static char *FileName;
int			m, n, nnz = 0;
double  	*a, *rhs;
int     	*rowind, *colptr; /* Used in Compressed Column Format (CCF). */
/*** Globals ***/


static int  ReadMatrixFromFile();
void		dPrint_Solution(char *what, SuperMatrix *A, int Len);


main(int argc, char *argv[])
{
/*
 * Purpose
 * =======
 * 
 * This is the small 5x5 erowptrmple used in the Sections 1 and 2 of the 
 * User's Guide to illustrate how to call a SuperLU routine, and the
 * matrix data structures used by SuperLU.
 *
 */

    SuperMatrix A, L, U, B;
    double   s, u, p, e, r, l;
    int      *perm_r; /* row permutations from partial pivoting */
    int      *perm_c; /* column permutation vector */
    int      nrhs = 1, info, i, permc_spec;

	/* Newly added variables */
	int		Error;
	double 	StartTime, SolveTime;

	/* For debugging */
	int		panel_size, relax;
	int		*etree, lwork = 0;
	double	*utime, t1;
	double  diag_pivot_thresh = 1.0, drop_tol = 0;
	extern SuperLUStat_t  SuperLUStat;
	SuperMatrix  AC;
	char	refact[1], trans[1];

	/* Read matrix file. */
	if (argc == 1) {
		fprintf(stderr, "Error: No data file.\nUsage: superlu data_file.\n");
		exit(1);
	}
	else 
		FileName = *(++argv);

	Error = ReadMatrixFromFile();

	if (Error) {
		fprintf(stderr, "Error: Failed reading data file. \n ");
		exit(1);
	}

	/* Allocate memory for Compressed Column Format. */

	/* Not needed !!
    if ( !(at = doubleMalloc(nnz)) ) ABORT("Malloc fails for at[].");
    if ( !(rowind = intMalloc(nnz)) ) ABORT("Malloc fails for rowind[].");
    if ( !(colptr = intMalloc(n+1)) ) ABORT("Malloc fails for colptr[].");
	*/

	/* Print to test */
	/*
	printf("\n at: \n");
	for (i = 0; i < nnz; i++)   printf(" %.2lf ", at[i]);
	printf("\n rowind : \n");
	for (i = 0; i < nnz; i++)   printf(" %d ", rowind[i]);
	printf("\n colptr: \n");
	for (i = 0; i < n+1; i++)   printf(" %d ", colptr[i]);
	*/
	printf("\n");

	/*printf("\n"); exit(1);*/ 
	/* Terminate the program here. */


    /* Create matrix A in the format expected by SuperLU. */
	/* Note: Must use 'at' */
    dCreate_CompCol_Matrix(&A, m, n, nnz, a, rowind, colptr, NC, D_, GE);
    
    /* Create right-hand side matrix B. */

    dCreate_Dense_Matrix(&B, m, nrhs, rhs, m, DN, D_, GE);

    if ( !(perm_r = intMalloc(m)) ) ABORT("Malloc fails for perm_r[].");
    if ( !(perm_c = intMalloc(n)) ) ABORT("Malloc fails for perm_c[].");

    /*
     * Get column permutation vector perm_c[], according to permc_spec:
     *   permc_spec = 0: use the natural ordering 
     *   permc_spec = 1: use minimum degree ordering on structure of A'*A
     *   permc_spec = 2: use minimum degree ordering on structure of A'+A
     */    	
    /*permc_spec = 1;*/   /* Options: 0, 1, 2 */
	do {
		printf("Choose ordering option ( 1 -- A'*A; 2 -- A'+A ) : ");
		scanf("%d", &permc_spec);
	} while ( permc_spec != 1 && permc_spec != 2 );
	putchar('\n');
	/*printf("permc_spec = %d\n", permc_spec);*/

	fprintf(stderr, "Get column permutation ... \n");   fflush(stderr);
    get_perm_c(permc_spec, &A, perm_c);

	/* Solve and record time. */
	printf("Start solving ... \n");
	fflush(stderr);

	StartTime = Time();

#if	DEBUG
	*refact = 'N';
	*trans = 'N';
	panel_size = sp_ienv(1);
	relax = sp_ienv(2);
	StatInit(panel_size, relax);
	utime = SuperLUStat.utime;
	etree = intMalloc(A.ncol);

	printf("- Preordering ... \n");	fflush(stdout);
    t1 = SuperLU_timer_();

	/**************** BUG ***************************/
    /*sp_preorder(refact, &A, perm_c, etree, &AC);*/
    xx_preorder(refact, &A, perm_c, etree, &AC);
	/************************************************/

    utime[ETREE] = SuperLU_timer_() - t1;

    printf("Factor PA = LU ... relax %d\tw %d\tmaxsuper %d\trowblk %d\n",
      relax, panel_size, sp_ienv(3), sp_ienv(4));

	printf("- Factoring ... \n");	   fflush(stdout);
    t1 = SuperLU_timer_();
    /* Compute the LU factorization of A. */
    dgstrf(refact, &AC, diag_pivot_thresh, drop_tol, relax, panel_size,
       etree, NULL, lwork, perm_r, perm_c, &L, &U, &info);
    utime[FACT] = SuperLU_timer_() - t1;

	printf("- Solving ... \n");	   fflush(stdout);
    t1 = SuperLU_timer_();
    if ( info == 0 ) {
        /* Solve the system A*X=B, overwriting B with X. */
        dgstrs (trans, &L, &U, perm_r, perm_c, &B, &info);
    }
    utime[SOLVE] = SuperLU_timer_() - t1;

    SUPERLU_FREE (etree);
    Destroy_CompCol_Permuted(&AC);

    PrintStat( &SuperLUStat );
    StatFree();
#else
    dgssv(&A, perm_c, perm_r, &L, &U, &B, &info);
#endif

	SolveTime = Time() - StartTime;

	/* Check the 'info'. */
    printf("\ninfo = %d ", info);   /* info = 0 : successful exit */
    if (info == 0)  printf(": Successful exit. \n");
    else            printf(": Unsuccessful exit. \n");

	/* Print the time used in solving. */
	printf("\n* Factor and Solve time = %.3lf seconds.\n", SolveTime);

	/* Print the solution which is in 'B'. */
	dPrint_Solution("Truncated Solution:", &B, 10);
	
	/*
    dPrint_Dense_Matrix("B", &B);
    dPrint_CompCol_Matrix("A", &A);
    dPrint_CompCol_Matrix("U", &U);
    dPrint_SuperNode_Matrix("L", &L);
    print_int_vec("\nperm_r", m, perm_r);
	*/	

    /* De-allocate storage */
    SUPERLU_FREE (a);
    SUPERLU_FREE (rowind);
    SUPERLU_FREE (colptr);
	/* Above three lines newly added. */

    SUPERLU_FREE (rhs);
    SUPERLU_FREE (perm_r);
    SUPERLU_FREE (perm_c);
    Destroy_CompCol_Matrix(&A);
    Destroy_SuperMatrix_Store(&B);
    Destroy_SuperNode_Matrix(&L);
    Destroy_CompCol_Matrix(&U);
}


static int  ReadMatrixFromFile()
{
	FILE  	*pMatrixFile;
	int   	Error, Row, Col, Count = 0, LineNumber, RHS_Col;
	int	  	Size = 0;
	long  	Reads;
	double 	Real;
	char  	sInput[BUFSIZ], sType[BUFSIZ];
	static char Message[BUFSIZ];

	static char *EndOfFile = "Unexpected end of file '%s' at line %d.\n";
	static char *Syntax = "Syntax error in file '%s' at line %d.\n";
	static char *Index = "Index error in file '%s' at line %d.\n";

	int		i, j;    
	/* Indices used to trace the three arrays in Compressed Row Format. */
	int		curr_col;		/* The current row index. */
	int		M;				/* For printing test. */

	pMatrixFile = fopen(FileName, "r");
	if (pMatrixFile == NULL) {
		fprintf(stderr, "Failed to open file: %s \n", FileName);
		return 1;
	}
	
	LineNumber = 1;

	/* Read and print label. */
	if (NULL == fgets(Message, BUFSIZ, pMatrixFile)) {
		fprintf(stderr, EndOfFile, FileName, LineNumber);
		return 1;
	}

	/* Check by print */
	/* printf(" Message = %s \n", Message); */

	/* Read the size of matrix. */
	LineNumber++;

	if (NULL == fgets(sInput, BUFSIZ, pMatrixFile)) {
		fprintf(stderr, EndOfFile, FileName, LineNumber);
		return 1;
	}
	if ((Reads = sscanf(sInput, "%d %s", &Size, sType)) < 1) {
		fprintf(stderr, Syntax, FileName, LineNumber);
		return 1;
	}

	/* Print to Check */
	/* printf(" Reads = %d \n", Reads); */

	/* Set the size of matrix A. (m, n) are globals. */
	m = n = Size;

	/* Read data file once to get the nonzero count (nnz). */
	nnz = 0;	/* Initialize num_of_nonzeros */
	do {
		LineNumber++;

		if (NULL == fgets(sInput, BUFSIZ, pMatrixFile)) {
			fprintf(stderr, EndOfFile, FileName, LineNumber);
			return 1;
		}

		Reads = sscanf(sInput, "%d%d%lf", &Row, &Col, &Real);

		if (Reads != 3) {
			fprintf(stderr, Syntax, FileName, LineNumber);
			return 1;
		}
		if (Row < 0 || Row > Size || Col < 0 || Col > Size) {
			fprintf(stderr, Index, FileName, LineNumber);
			return 1;
		}

		/* Print to check */
		/* printf(" %d, %d, %.2lf \n", Row, Col, Real); */

		/* Get the total num of nonzeros so that memory can be allocated. */
		if (Row > 0 && Col > 0)		++nnz;

	} while (Row != 0 && Col != 0);

	/* Print to check */
	printf("\nSize = %d,  Type = %s \n", Size, sType);
	printf("nnz = %d \n", nnz); 
	printf("sparsity (nnz/n) = %.6lf \n", (double) nnz/Size ); 

	/* Allocate memory for Column Compressed Format (CCF). */

    if ( !(a = doubleMalloc(nnz)) ) ABORT("Malloc fails for a[].");
    if ( !(rowind = intMalloc(nnz)) ) ABORT("Malloc fails for colind[].");
    if ( !(colptr = intMalloc(n+1)) ) ABORT("Malloc fails for rowptr[].");

	/* Allocate memory for RHS. */

    if ( !(rhs = doubleMalloc(m)) ) ABORT("Malloc fails for rhs[].");


    /* Read data file again to set up arrays in Column Compressed Format. */

	/* Resume the file pointer */
	rewind(pMatrixFile);

	/* Skip the first two lines. */
	fgets(Message, BUFSIZ, pMatrixFile);
	fgets(Message, BUFSIZ, pMatrixFile);
	LineNumber = 2;

	LineNumber++;
    fgets(sInput, BUFSIZ, pMatrixFile);
    Reads = sscanf(sInput, "%d%d%lf", &Row, &Col, &Real);

	i = j = 0;
	curr_col = 0;		/* col index counts from 1 in the data file */

    while (Row > 0 && Col > 0) {

		a[i] = Real;			/* nonzero value */
		rowind[i] = Row - 1;	/* rowind = row index */

		if (Col > curr_col) {
			colptr[j++] = i; 
			curr_col++;
		}
		i++;

		LineNumber++;
        fgets(sInput, BUFSIZ, pMatrixFile);
        Reads = sscanf(sInput, "%d%d%lf", &Row, &Col, &Real);
    } 

	colptr[j] = i;		/* Append the ending index in rowptr. */

	/* Print to test */
	/*	
	M = (nnz > 20)? 20 : nnz;
	printf("\na: \n");
	for (i = 0; i < M; i++)   printf(" %lf ", a[i]);
	printf("\nrowind : \n");
	for (i = 0; i < M; i++)   printf(" %d ", rowind[i]);
	printf("\ncolptr: \n");
	for (i = 0; i < M; i++)   printf(" %d ", colptr[i]);
	*/	

	/* Continue to set up the RHS */
	for (i = 0; i < Size; i++) {
		LineNumber++;

		if (NULL == fgets(sInput, BUFSIZ, pMatrixFile)) {
			fprintf(stderr, EndOfFile, FileName, LineNumber);
			return 1;
		}

		Reads = sscanf(sInput, "%lf", &Real);

		if (Reads != 1) {
			fprintf(stderr, Syntax, FileName, LineNumber);
			return 1;
		}

		rhs[i] = Real;
	}

	/* Print to test */
	/*
	printf("\n rhs: \n");
	for (i = 0; i < m; i++)   printf(" %.2lf ", rhs[i]);
	*/


	(void) fclose(pMatrixFile);

	printf("\n");

	return 0;		/* No error. */
}

void	dPrint_Solution(char *what, SuperMatrix *A, int Len)
{
    DNformat     *Astore;
    register int i;
    double       *dp;

    printf("\n%s\n", what);
    Astore = (DNformat *) A->Store;
    dp = (double *) Astore->nzval;
    for (i = 0; i < Len; ++i) 
		printf("%lg \n", dp[i]);
    printf("\n");
    fflush(stdout);
}


void   xx_preorder(char *refact,  SuperMatrix *A, int *perm_c, 
	    	int *etree, SuperMatrix *AC)
{
    NCformat  *Astore;
    NCPformat *ACstore;
    int       *iwork, *post;
    register  int n, i;

    n = A->ncol;
    iwork = (int*) SUPERLU_MALLOC((n+1)*sizeof(int)); 
    if ( !iwork ) ABORT("SUPERLU_MALLOC fails for iwork[]");
    
    /* Apply column permutation perm_c to A's column pointers so to
       obtain NCP format in AC = A*Pc.  */
    AC->Stype       = NCP;
    AC->Dtype       = A->Dtype;
    AC->Mtype       = A->Mtype;
    AC->nrow        = A->nrow;
    AC->ncol        = A->ncol;
    Astore          = A->Store;
    ACstore = AC->Store = (void *) SUPERLU_MALLOC( sizeof(NCPformat) );
    if ( !ACstore ) ABORT("SUPERLU_MALLOC fails for ACstore");
    ACstore->nnz    = Astore->nnz;
    ACstore->nzval  = Astore->nzval;
    ACstore->rowind = Astore->rowind;
    ACstore->colbeg = (int*) SUPERLU_MALLOC(n*sizeof(int));
    if ( !(ACstore->colbeg) ) ABORT("SUPERLU_MALLOC fails for ACstore->colbeg");
    ACstore->colend = (int*) SUPERLU_MALLOC(n*sizeof(int));
    if ( !(ACstore->colend) ) ABORT("SUPERLU_MALLOC fails for ACstore->colend");

#ifdef DEBUG
    /*print_int_vec("pre_order:", n, perm_c);*/

	printf("--- Check 'perm_c' ... \n");  fflush(stdout);
    check_perm("Initial perm_c", n, perm_c);
#endif      

    for (i = 0; i < n; i++) {
	ACstore->colbeg[perm_c[i]] = Astore->colptr[i]; 
	ACstore->colend[perm_c[i]] = Astore->colptr[i+1];
    }
	
    if ( lsame_(refact, "N") ) {
	
	/* Compute the column elimination tree */

	printf("--- Compute etree ... \n");  fflush(stdout);
	sp_coletree(ACstore->colbeg, ACstore->colend, ACstore->rowind,
		    A->nrow, A->ncol, etree);
#ifdef DEBUG	
	print_int_vec("etree:", n, etree);
#endif	
	
	/* Post order etree */
	printf("*** Get post order etree ... \n");  fflush(stdout);

	post = (int *) xxTreePostorder(n, etree);	 /** <<< BUG !! >>> **/

	/* TreePostorder() defined in .../SRC/sp_coletree.c */

	/* for (i = 0; i < n+1; ++i) inv_post[post[i]] = i;
	   iwork = post; */

#ifdef DEBUG
	print_int_vec("post:", n+1, post);
	printf("--- Check 'post' ... \n");  fflush(stdout);
	check_perm("post", n, post);	
#endif	

	/* Renumber etree in postorder */
	for (i = 0; i < n; ++i) iwork[post[i]] = post[etree[i]];
	for (i = 0; i < n; ++i) etree[i] = iwork[i];

#ifdef DEBUG	
	/*print_int_vec("postorder etree:", n, etree);*/
#endif
	
	/* Postmultiply A*Pc by post[] */
	for (i = 0; i < n; ++i) iwork[post[i]] = ACstore->colbeg[i];
	for (i = 0; i < n; ++i) ACstore->colbeg[i] = iwork[i];
	for (i = 0; i < n; ++i) iwork[post[i]] = ACstore->colend[i];
	for (i = 0; i < n; ++i) ACstore->colend[i] = iwork[i];

	for (i = 0; i < n; ++i)
	    iwork[i] = post[perm_c[i]];  /* product of perm_c and post */
	for (i = 0; i < n; ++i) perm_c[i] = iwork[i];

#ifdef DEBUG
	/*print_int_vec("Pc*post:", n, perm_c);*/
	printf("--- Check final 'perm_c' ... \n");  fflush(stdout);
	check_perm("final perm_c", n, perm_c);	
#endif

	SUPERLU_FREE (post);

    } /* if refact = 'N' */

    SUPERLU_FREE (iwork);

}


check_perm(char *what, int n, int *perm)
{
    register int i;
    int          *marker;

    marker = (int *) calloc(n, sizeof(int));   /* Initialized to zero */

    /* marker = (int*) SUPERLU_MALLOC(n*sizeof(int)); */
    if ( !marker ) ABORT("'calloc' failed for marker[]");

	/* Initialize marker[] to all zero. */
	/*for (i = 0; i < n; ++i) 	marker[i] = 0;*/

    for (i = 0; i < n; ++i) {
		if ( marker[perm[i]] == 1 || perm[i] >= n ) {
	    	printf("%s: Not a valid PERM[%d] = %d\n", what, i, perm[i]);
	    	ABORT("check_perm");
		} 
		else {
	    	marker[perm[i]] = 1;
		}
    }

    SUPERLU_FREE(marker);
    return 0;
}


int  *first_kid, *next_kid;  /* Linked list of children. */
int  *post, postnum;
/*
 * Depth-first search from vertex v.
 */
void etdfs ( int v, int n )
{
    int w;
/*
	if ( v > n || (w = first_kid[v]) > n) {
		fprintf(stderr, "\nError: Out of range, v = %d, w = %d.\n", v, w);
		fflush(stderr);
		exit(1);
	}
	*/
/*
    for (w = first_kid[v]; w != -1; w = next_kid[w]) {
		 printf("w:%d and v :%d \n",w,v); 
        etdfs (w, n);
    }
*/

   assert(v>=0 && v<=n);
	w=first_kid[v];
	for(;;)
	{
	if(w!=-1)
	{
	  etdfs(w,n);

	assert(w>=0 && w<=n);
	w=next_kid[w];
	}else
	break;
	}


 /* post[postnum++] = v;     in Matlab */

    post[v] = postnum++;     /* Modified by X.Li on 2/14/95 */
}


/*
 * Post order a tree
 */
int *xxTreePostorder( int n, int *parent)
{
    int v, dad;

    /* Allocate storage for working arrays and results  */
    first_kid =  intMalloc(n+1);
    next_kid  =  intMalloc(n+1);
    post      =  intMalloc(n+1);


    /* Set up structure describing children */
    for (v = 0; v <= n; first_kid[v++] = -1);
    for (v = n-1; v >= 0; v--) {
        dad = parent[v];
        next_kid[v] = first_kid[dad];
        first_kid[dad] = v;
    }

    /* Depth-first search from dummy root vertex #n */
    postnum = 0;
    etdfs (n, n);

    SUPERLU_FREE (first_kid);
    SUPERLU_FREE (next_kid);
    return post;
}


