/************************************************************************
	superlu_rcf.c  -- Row-Compressed-Format

	Copied from ../EXAMPLE/superlu.c
	Modified by G. Shi, 10-23-02, for testing some circuit systems.
	Data (A,b) are provided by files in the format of (row, col, value).
*************************************************************************/

/*
 * -- SuperLU routine (version 2.0) --
 * Univ. of California Berkeley, Xerox Palo Alto Research Center,
 * and Lawrence Berkeley National Lab.
 * November 15, 1997
 *
 */
#include "dsp_defs.h"
#include "util.h"
#include "time.h"


#define		Init_Index		0

#define     HZ      CLK_TCK

double  Time()
{
    struct time {long  user, system, childuser, childsystem;} time;

    (void) times(&time);
    return  (double) time.user / (double) HZ;
}


static char *FileName;

int		m, n, nnz = 0;
double  *a, *rhs;
int     *colind, *rowptr; /* Used in Compressed Row Format (CRF). */


static int  ReadMatrixFromFile();
void        dPrint_Solution(char *what, SuperMatrix *A, int Len);


main(int argc, char *argv[])
{
/*
 * Purpose
 * =======
 * 
 * This is the small 5x5 erowptrmple used in the Sections 1 and 2 of the 
 * User's Guide to illustrate how to call a SuperLU routine, and the
 * matrix data structures used by SuperLU.
 *
 */

    SuperMatrix A, L, U, B;
    double   s, u, p, e, r, l;
    int      *perm_r; /* row permutations from partial pivoting */
    int      *perm_c; /* column permutation vector */
    int      nrhs = 1, info, i, permc_spec;

	/* Newly added variables */
	int		Error;
	double  StartTime, SolveTime, Time0;
	double  *at;		/* Stores the transpose of 'a' */
	int     *rowind, *colptr;  /* Used in Compressed Column Format (CCF)

	
	/* Get the beginning time for measuring the time duration. */
	Time0 = Time();	

	/* Read matrix file. */
	if (argc == 1) {
		fprintf(stderr, "Error: No data file.\nUsage: superlu data_file.\n");
		exit(1);
	}
	else 
		FileName = *(++argv);

	Error = ReadMatrixFromFile();

	if (Error) {
		fprintf(stderr, "Error: Failed reading data file. \n ");
		exit(1);
	}

	/* Allocate memory for Compressed Column Format. */

	/* Not needed !!
    if ( !(at = doubleMalloc(nnz)) ) ABORT("Malloc fails for at[].");
    if ( !(rowind = intMalloc(nnz)) ) ABORT("Malloc fails for rowind[].");
    if ( !(colptr = intMalloc(n+1)) ) ABORT("Malloc fails for colptr[].");
	*/

	/* Convert from Compressed Row Format to Compressed Column Format. */
	dCompRow_to_CompCol(m, n, nnz, a, colind, rowptr, &at, &rowind, &colptr);
	/* This function is defined in '../SRC/dutil.c and dutil.h'. */

	/* Free memory used by CRF: 'a, colind, rowptr' */
	free(a);
	free(colind);
	free(rowptr);

	/* Print to test */
	/*	
	printf("\nat: \n");
	for (i = 0; i < nnz; i++)   printf(" %g ", at[i]);
	printf("\nrowind : \n");
	for (i = 0; i < nnz; i++)   printf(" %d ", rowind[i]);
	printf("\ncolptr: \n");
	for (i = 0; i < n+1; i++)   printf(" %d ", colptr[i]);
	printf("\nrhs: \n");
	for (i = 0; i < n; i++)   printf(" %g ", rhs[i]);
	*/	
	printf("\n");

	/* printf("\n"); exit(1); */	
	/* Terminate the program here. */


    /* Create matrix A in the format expected by SuperLU. */
	/* Note: Must use 'at' */
    dCreate_CompCol_Matrix(&A, m, n, nnz, at, rowind, colptr, NC, D_, GE);
    
    /* Create right-hand side matrix B. */

    dCreate_Dense_Matrix(&B, m, nrhs, rhs, m, DN, D_, GE);

    if ( !(perm_r = intMalloc(m)) ) ABORT("Malloc fails for perm_r[].");
    if ( !(perm_c = intMalloc(n)) ) ABORT("Malloc fails for perm_c[].");

    /*
     * Get column permutation vector perm_c[], according to permc_spec:
     *   permc_spec = 0: use the natural ordering 
     *   permc_spec = 1: use minimum degree ordering on structure of A'*A
     *   permc_spec = 2: use minimum degree ordering on structure of A'+A
     */    	
    permc_spec = 1;   /* Options: 0, 1, 2 */

    get_perm_c(permc_spec, &A, perm_c);

    StartTime = Time();
    dgssv(&A, perm_c, perm_r, &L, &U, &B, &info);
	SolveTime = Time() - StartTime;
    
	/* Check the 'info'. */
	printf("\ninfo = %d ", info);   /* info = 0 : successful exit */
	if (info == 0)  printf(": Successful exit. \n");
	else            printf(": Unsuccessful exit. \n");

    /* Print the time used in solving. */
	putchar('\n');
    printf("* Factor and Solve time = %.3lf seconds (Option %d).\n", \
            SolveTime, permc_spec);
	
	printf("* Total time used = %.3lf seconds.\n", Time() - Time0);

	/* Print the solution which is in 'B'. */
    dPrint_Solution("Truncated Solution:", &B, 9);
	
	/*
    dPrint_Dense_Matrix("B", &B);
    dPrint_CompCol_Matrix("A", &A);
    dPrint_CompCol_Matrix("U", &U);
    dPrint_SuperNode_Matrix("L", &L);
    print_int_vec("\nperm_r", m, perm_r);
	*/	

    /* De-allocate storage */
    SUPERLU_FREE (at);
    SUPERLU_FREE (rowind);
    SUPERLU_FREE (colptr);
	/* Above three lines newly added. */

    SUPERLU_FREE (rhs);
    SUPERLU_FREE (perm_r);
    SUPERLU_FREE (perm_c);
    Destroy_CompCol_Matrix(&A);
    Destroy_SuperMatrix_Store(&B);
    Destroy_SuperNode_Matrix(&L);
    Destroy_CompCol_Matrix(&U);
}


static int  ReadMatrixFromFile()
{
	FILE  	*pMatrixFile;
	int   	Error, Row, Col, Count = 0, LineNumber, RHS_Col;
	int	  	Size = 0;
	long  	Reads;
	double 	Real;
	char  	sInput[BUFSIZ], sType[BUFSIZ];
	static char Message[BUFSIZ];

	static char *EndOfFile = "Unexpected end of file '%s' at line %d.\n";
	static char *Syntax = "Syntax error in file '%s' at line %d.\n";
	static char *Index = "Index error in file '%s' at line %d.\n";

	int		i, j;    
	/* Indices used to trace the three arrays in Compressed Row Format. */
	int		curr_row;		/* The current row index. */

	pMatrixFile = fopen(FileName, "r");
	if (pMatrixFile == NULL) {
		fprintf(stderr, "Failed to open file: %s \n", FileName);
		return 1;
	}
	
	LineNumber = 1;

	/* Read and print label. */
	if (NULL == fgets(Message, BUFSIZ, pMatrixFile)) {
		fprintf(stderr, EndOfFile, FileName, LineNumber);
		return 1;
	}

	/* Check by print */
	/* printf(" Message = %s \n", Message); */

	/* Read the size of matrix. */
	LineNumber++;

	if (NULL == fgets(sInput, BUFSIZ, pMatrixFile)) {
		fprintf(stderr, EndOfFile, FileName, LineNumber);
		return 1;
	}
	if ((Reads = sscanf(sInput, "%d %s", &Size, sType)) < 1) {
		fprintf(stderr, Syntax, FileName, LineNumber);
		return 1;
	}

	/* Print to Check */
	/* printf(" Reads = %d \n", Reads); */

	/* Set the size of matrix A. (m, n) are globals. */
	m = n = Size;

	/* Read data file once to get the nonzero count (nnz). */
	nnz = 0;	/* Initialize num_of_nonzeros */
	do {
		LineNumber++;

		if (NULL == fgets(sInput, BUFSIZ, pMatrixFile)) {
			fprintf(stderr, EndOfFile, FileName, LineNumber);
			return 1;
		}

		Reads = sscanf(sInput, "%d%d%lf", &Row, &Col, &Real);

		if (Reads != 3) {
			fprintf(stderr, Syntax, FileName, LineNumber);
			return 1;
		}
		if (Row < 0 || Row > Size || Col < 0 || Col > Size) {
			fprintf(stderr, Index, FileName, LineNumber);
			return 1;
		}

		/* Print to check */
		/* printf(" %d, %d, %.2lf \n", Row, Col, Real); */

		/* Get the total num of nonzeros so that memory can be allocated. */
		if (Row > 0 && Col > 0)		++nnz;

	} while (Row != 0 && Col != 0);

	/* Print to check */
	printf("\nSize = %d,  Type = %s \n", Size, sType);
	printf("nnz = %d \n", nnz); 
	printf("sparsity (n/nnz) = %.6f \n", (float) nnz/(float) Size ); 

	/* Allocate memory for Compressed Row Format. */

    if ( !(a = doubleMalloc(nnz)) ) ABORT("Malloc fails for a[].");
    if ( !(colind = intMalloc(nnz)) ) ABORT("Malloc fails for colind[].");
    if ( !(rowptr = intMalloc(n+1)) ) ABORT("Malloc fails for rowptr[].");

	/* Allocate memory for RHS. */

    if ( !(rhs = doubleMalloc(m)) ) ABORT("Malloc fails for rhs[].");


    /* Read data file again to set up arrays in Compressed Row Format. */

	/* Resume the file pointer */
	rewind(pMatrixFile);

	/* Skip the first two lines. */
	fgets(Message, BUFSIZ, pMatrixFile);
	fgets(Message, BUFSIZ, pMatrixFile);
	LineNumber = 2;

	LineNumber++;
    fgets(sInput, BUFSIZ, pMatrixFile);
    Reads = sscanf(sInput, "%d%d%lf", &Row, &Col, &Real);

	i = j = 0;
	curr_row = 0;		/* row index counts from 1 in the data file */

    while (Row > 0 && Col > 0) {

		a[i] = Real;			/* nonzero value */
		colind[i] = Col - 1;	/* colind = column index */

		if (Row > curr_row) {
			rowptr[j++] = i; 
			curr_row++;
		}
		i++;

		LineNumber++;
        fgets(sInput, BUFSIZ, pMatrixFile);
        Reads = sscanf(sInput, "%d%d%lf", &Row, &Col, &Real);
    } 

	rowptr[j] = i;		/* Append the ending index in rowptr. */

	/* Print to test */
	/*
	printf("\n a: \n");
	for (i = 0; i < nnz; i++)   printf(" %.2lf ", a[i]);
	printf("\n colind : \n");
	for (i = 0; i < nnz; i++)   printf(" %d ", colind[i]);
	printf("\n rowptr: \n");
	for (i = 0; i < n+1; i++)   printf(" %d ", rowptr[i]);
	*/

	/* Continue to set up the RHS */
	for (i = 0; i < Size; i++) {
		LineNumber++;

		if (NULL == fgets(sInput, BUFSIZ, pMatrixFile)) {
			fprintf(stderr, EndOfFile, FileName, LineNumber);
			return 1;
		}

		Reads = sscanf(sInput, "%lf", &Real);

		if (Reads != 1) {
			fprintf(stderr, Syntax, FileName, LineNumber);
			return 1;
		}

		rhs[i] = Real;
	}

	/* Print to test */
	/*
	printf("\n rhs: \n");
	for (i = 0; i < m; i++)   printf(" %.2lf ", rhs[i]);
	*/


	(void) fclose(pMatrixFile);

	printf("\n");

	return 0;		/* No error. */
}

void    dPrint_Solution(char *what, SuperMatrix *A, int Len)
{
    DNformat     *Astore;
    register int i;
    double       *dp;

    printf("\n%s\n", what);
    Astore = (DNformat *) A->Store;
    dp = (double *) Astore->nzval;
    for (i = 0; i < Len; ++i)
        printf("%lg \n", dp[i]);
    printf("\n");
    fflush(stdout);
}

